/*
Copyright © 2013 Alexander Kobzar (sashakobzar@gmail.com)

GETALP
Grenoble Informatics Laboratory
France

This script allows you to filter a Blast file based on a number of criteria.
There are three types of fiters: annotation, max lenght, and word-based.
The word-based one is aimed at dealing with phrasal verbs in split word order, which are usually incorrectly indentified by MWEtoolkit.
By means of the annotation filter, sentences can be filtered out depending on whether a sentence is annotated with either a particular item, level, or both.
Finally, you can specify a maximum sentence length so as to filter out all sentences whose wordcount exceeds the threshold.
 */
using System;
using System.IO; 
using System.Text.RegularExpressions;   
using System.Linq;
using System.Collections.Generic ;

namespace Cleanblast
{
	class MainClass
	{
		
		public static void Main (string[] args)
		{
			string blastFile, filteredBlast, excludeAnnot = "", sentences = "", blastAln = "";
			short maxLength = 0, parseMwePos;
			string [] blastSourceWords;
			short [] mwePositions;
			int wordFiltExcl = 0, maxLengthExcl = 0, annotExcl = 0, sentIncl = 0;
			bool wordFilters = true;
			StreamReader readBlast;
			StreamWriter writeFiltBlast;
			string message = @"filterblast.exe <Blast file> <filtered Blast file> [OPTIONS]

Options:
--------
-exclude <[level][-item]> - filter out sentences having the specified annotation

-maxlenght <number> - sentences whose word count exceeds the threshold will be filtered out

-diswordfilt - disable word-based filters for phrasal verbs in split word order"; 
			#region Process command line arguments
			if (args.Count () == 0) {
				Console.WriteLine ("You did not specify all required parameters.\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' or  for help.");
				return;
			} else if (args [0] == "/?" || args [0] == "-help") {
				Console.WriteLine (message); 
				return;			
			}
			if (!File.Exists (args [0])) {
				Console.WriteLine ("{0} does not exist.", args [0]);
				return;
			}
			//Process options
			for (int i=2; i< args.Count (); i++) {
				switch (args [i]) {
				case "-exclude":
					if (i + 1 == args.Count ()) {
						Console.WriteLine ("You have to specify a value for the -exclude option in the following format: [level][-item]");
						return;
					} else if (!Regex.IsMatch (args [i + 1], "[^-]?(-[^-])?") && args [i + 1] != "") {
						Console.WriteLine ("The annotation should be in the following format: [level][-item].");
						return;
					}
					excludeAnnot = args [i + 1];
					i++;
					break;
				case "-maxlength":
					if (i + 1 == args.Count ()) {
						Console.WriteLine ("You have to specify a threshold for the -maxlength option.");
						return;
					} else {
						short parseRes;
						if (!short.TryParse (args [i + 1], out parseRes) || short.Parse (args [i + 1]) <= 0) {
							Console.WriteLine ("The max lenght should be a positive integer.");
							return;
						}
					}
					maxLength = short.Parse (args [i + 1]);
					i++;
					break;
				case "-diswordfilt":
					wordFilters = false;
					break;
				default :
					Console.WriteLine ((args [i].StartsWith ("-") ? "Invalid option: " : "Invalid argument: ") + args [i] + "\r\nType '[mono] filterblast.exe -help' or '[mono] filterblast.exe /?' for help."); 
					return;
					break;
				}
			}
			#endregion 
			blastFile = args [0];
			filteredBlast = args [1] + ".blast";
			readBlast = new StreamReader (blastFile);
			try {
				writeFiltBlast = new StreamWriter (filteredBlast);
			} catch (Exception e) {
				Console.WriteLine ("The filtered Blast file you specified, {0}, cannot be created due to the following reason:\r\n{1}", args [1], e.Message);
				return;
			}
			//Copy the header
			writeFiltBlast.WriteLine (readBlast.ReadLine ());
			writeFiltBlast.WriteLine (readBlast.ReadLine ());	
			//Read a Blast file
			while (readBlast.Peek ()!=-1) {
				//Store sentence information except alignment
				sentences = readBlast.ReadLine () + "\r\n";
				blastSourceWords = sentences.ToLower ().Split (new char []{' '}); 
				sentences += readBlast.ReadLine () + "\r\n";
				sentences += readBlast.ReadLine () + "\r\n";
				sentences += readBlast.ReadLine () + "\r\n";
				if (readBlast.Peek () == -1) {
					Console.WriteLine ("{0} is malformed. Inspect the end of the file and check whether headers are not missing.", blastFile);
					readBlast.Close ();
					writeFiltBlast.Close ();
					return;
				}
				//Store alignment
				blastAln = readBlast.ReadLine ();
				//Check the sentence length
				if (blastSourceWords.Length <= maxLength || maxLength == 0) {
					//Treat each annotation of a sentence
					foreach (string annotation in blastAln.Split (new char []{' '})) {
						if (annotation == "")
							continue;
						if (!annotation.Contains ("#")) {
							Console.WriteLine ("The following line in {0} is malformed:\r\n{1}", blastFile, blastAln);
							return;
						}
						mwePositions = annotation.Substring (0, annotation.IndexOf ('#')).Split (new char []{','}).Select (x => {
							if (!short.TryParse (x, out parseMwePos)) {
								return (short)-2;
							}
							return parseMwePos;}
						).ToArray ();
						#region Inspect alignment information
						if (mwePositions.Contains ((short)-2)) {
							Console.WriteLine ("The following line in {0} contains an error:\r\n{1}", blastFile, blastAln);
							readBlast.Close ();
							writeFiltBlast.Close ();
							return;
						}
						#endregion 
						//Check whether the sentence should not be included based on the user-defined level and item annotation.
						if (excludeAnnot != "" && ((excludeAnnot.Contains ("-") && annotation.EndsWith ("#" + excludeAnnot)) || (!excludeAnnot.Contains ("-") && annotation.Substring (annotation.LastIndexOf ('#') + 1).StartsWith (excludeAnnot + "-")) || (excludeAnnot.StartsWith ("-") && annotation.EndsWith (excludeAnnot)))) {
							annotExcl ++;
							continue;
						}
						#region Word-based filters
						if (wordFilters) {
							//Verbs be, go, walk, do, see used with down,up,in,out
							if ((new string []{"be","go","walk","do","see"}.Contains (blastSourceWords [mwePositions [0]])) && (new string [] {
							"down",
							"up",
							"in",
							"out"
						}.Contains (blastSourceWords [mwePositions [1]]))) {	
								wordFiltExcl ++;
								continue;
							}			    
							//Particles about, well, at
							if ((new string []{"about","well","at"}.Contains (blastSourceWords [mwePositions [1]]))) {	
								wordFiltExcl ++;
								continue;
							}
							//Any particle followed by locative words "here" and "there"
							if (mwePositions [1] != blastSourceWords.Count () - 1 && new string [] {
								"there",
								"here"
							}.Contains (blastSourceWords [mwePositions [1] + 1])) { 
								wordFiltExcl ++;
								continue;
							}
							//There is a particle "on, down, up, in, back" between the verb and the particle
							if (blastSourceWords.Skip (mwePositions [0] + 1).Take (mwePositions [1] - mwePositions [0] - 1).Intersect (new string [] {
							"on",
							"down",
							"up",
							"in",
							"back"
						}
							).Count () != 0) {	
								wordFiltExcl ++;
								continue;				
							}
							//Locative particles "up" and "down" are preceded by the word "way"
							if ((new string []{"up","down"}.Contains (blastSourceWords [mwePositions [1]])) && (blastSourceWords [mwePositions [1] - 1] == "way")) {
								wordFiltExcl ++;
								continue;
							}
							//Particle "down" is preceded by "upside" 
							if ((blastSourceWords [mwePositions [1]] == "down") && (blastSourceWords [mwePositions [1] - 1] == "upside")) {
								wordFiltExcl ++;
								continue;
							}
							//particle "out" is preceded by "inside"
							if ((blastSourceWords [mwePositions [1]] == "out") && (blastSourceWords [mwePositions [1] - 1] == "inside")) {
								wordFiltExcl ++;
								continue;
							}
							//Particle "over" is preceded by the word "all"
							if ((blastSourceWords [mwePositions [1]] == "over") && (blastSourceWords [mwePositions [1] - 1] == "all")) {
								wordFiltExcl ++;
								continue;
							}
						}
						#endregion 
						//Write the sentence which has passed all the filters
						writeFiltBlast.Write (sentences);
						writeFiltBlast.WriteLine (annotation + " ");
						sentIncl ++;
					}
				} else {
					maxLengthExcl ++;
				}
			}
			Console.WriteLine (filteredBlast + " is created."); 
			Console.WriteLine (sentIncl + " sentences to annotate."); 
			Console.WriteLine ("Total number of sentences fitered out: " + (maxLengthExcl + wordFiltExcl + annotExcl) +
				"\r\n--------------------------------------\r\nWord-based fiters: " + wordFiltExcl + "\r\nAnnotation filter: " + annotExcl + "\r\nMax lenghth filter: " + maxLengthExcl
			); 
			writeFiltBlast.Close ();
		} 
	}
}